# Variational search distributions

Prototype code implementing variational search distributions (VSD) and
experiments in the paper of the same name.

This code is given solely for the purposes of review, all code remains property
of the original authors and no permission is given for subsequent use,
modification or distribution.


## Installation

You can either use poetry or pip+venv for installation.

### Poetry

These instructions are for [poetry](https://python-poetry.org) virtual
environment and package management.

    cd path/to/variationalsearch
    poetry install
    poetry shell

Or to run the experiments,

    cd path/to/variationalsearch
    poetry install -E experiments
    git lfs install  # (not required for bundled code)
    git lfs pull  # (not required for bundled code)
    poetry shell

Now you should be able to use the code and run the experiments.


### Pip+Venv

Alternatively, you can use venv and pip, to install

    cd path/to/variationalsearch
    virtualenv --python="/path/to/python<3.12" vsdenv
    source vsdenv/bin/activate
    pip install .

Or to run the experiments, replace the last line above with,

    pip install .[experiments]
    git lfs install  # (not required for bundled code)
    git lfs pull  # (not required for bundled code)


## Experiments

### Biological sequence experiments

These instructions are for the _fitness landscape_ and _black-box optimization_
experiments in the variational search distributions paper. All experimental
configuration is contained in `experiments/config.py`.

Before running the experiments, if you do not have any pre-trained predictive
models in the `models/` directory, or if you wish to modify the preditive
models, you will need to run,

    train_cpe --dataset DATASET

to train a class probability estimator, or

    train_gp --dataset DATASET

for training a Gaussian process surrogate. See `train_cpe --help` and
`train_gp --help` for all options. We give the options for `DATASET` below.

- `GFP` for the partially synthetic (oracle predictor) avGFP dataset,
- `DHFR` for the complete-space DHFR dataset,
- `TFBIND8` or `TFBIND10` for the complete-space TF bind 8/10 datasets,
- `AAV` for the partially synthetic (oracle predictor) AAV dataset,
- `TRPB` for the complete-space TrpB dataset.

Some of these datasets also have a `_FL` prefix, e.g. `DHFR_FL`, which refers to
the "fitness landscape" version of an experiment. That is, where the fitness
threshold is fixed, and precision and recall are measured.

You may also wish to train a prior on some data, this can be done by calling the
`train_prior` command.

    train_prior --dataset DATASET

You can choose to train a prior on all the initial predictor training data, or
just the `fit` subset, see the `dataset["prior"]["use_threshold"]`
configuration parameter.

The experiments can be run by executing the shell script,

    ./run_seq_exp DATASET

Or, you can specify the dataset and method to run by calling,

    batch_rounds --dataset DATASET --method METHOD

See `batch_rounds --help` for all options and option values.

All surrogate model and experimental configuration can be viewed and changed in
`experiments/config.py`. Furthermore, an experiment will save its configuration
to a JSON file, which can also be modified and then used in the
`batch_rounds --config CONFIG` argument.

Run the following commands to replicate the results from the paper (you may
need to parallelize these on multiple GPUs). For the fitness landscape
experiments (section 4.2):

    run_seq_exp DHFR_FL
    run_seq_exp TRPB_FL
    run_seq_exp TFBIND8_FL

or to run the Gaussian process version in appendix section C.1, append the
`-gp` flag to the above command. For the black-box optimization experiments run
the following,

    run_seq_exp AAV
    run_seq_exp GFP

While the fitness landscape experiments can be run on a laptop, we run the BBO
experiments on a single NVIDIA H100 GPU equipped server. After these scripts
have completed, run the following command to create the plots we present in the
paper,

    plot_results --datadir DATASET

where `DATASET` is a dataset listed above. To run the ablation experiments, you
will need to modify the configuration file `AAV` or `GFP` configuration
dictionaries as we have commented in `experiments/config.py` and the run the
`batch_rounds` command explicitly for `--method VSD`.

We have also provided a [slurm](https://slurm.schedmd.com/sacct.html) run script
for these experiments, see `run_seq_exp_hpc.sh`.


### Handwritten digit generation

The handwritten digit generation experiment (section 4.1 from the paper) is
self-contained in the `experiments/scripts/digits.py` script, to run simply call

    digits

for running the digits experiments with the transformer variational
distribution, or append the `-lstm` flag to use the LSTM variational
distribution. NOTE: this will require a decent GPU with 32GB+ of memory to run
without modification -- we use an NVIDIA H100 GPU. All configuration for this
experiment is contained within the script.

We have also provided a [slurm](https://slurm.schedmd.com/sacct.html) run script
for these experiments, see `run_digits.sh`.


### Notebook

We have also included a toy experiment based on finding the super-level
distribution of a 2D mixture of Gaussians. This demonstrates VSD's ability to
also work in continuous spaces with only modification to the variational
distributions. See `notebooks/gmm_exploration_test.ipynb`.

This experiment was not featured in the paper.


## Acknowledgements
- TFBIND 8/10 data, https://huggingface.co/datasets/beckhamc/design_bench_data
- GFP and AAV data, https://github.com/kirjner/GGS
- DHFR data, supplementary material from https://doi.org/10.1126/science.adh3860
- TrpB data, supplementary material from https://doi.org/10.22002/h5rah-5z170.
